""" B{ImageCms} - PIL interface to little cms ICC Color Management System

License
=======

 - B{ImageCms} home page: U{http://trac.gbiloba.org/ImageCms}
 - Copyright (C) 2008 Frederic Mantegazza

This software is governed by the B{CeCILL-C} license under French law and
abiding by the rules of distribution of free software. You can use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
U{http://www.cecill.info}.

As a counterpart to the access to the source code and rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty and the software's author, the holder of the
economic rights, and the successive licensors have only limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading, using, modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean that it is complicated to manipulate, and that also
therefore means that it is reserved for developers and experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and, more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

History and needed modules
==========================

B{ImageCms} is based on pyCMS code:

 - pyCMS home page: U{http://www.cazabon.com}
 - pyCMS is Copyright (C) 2002-2003 Kevin Cazabon

The name has changed to better match other PIL modules. The license has also
changed, moved from LGPL to CeCILL-C to match french laws.

B{ImageCms} also uses little cms and PIL libraries:

 - little cms home page: U{http://www.littlecms.com}
 - little cms is Copyright (C) 1998-2008 Marti Maria
 - PIL home page: U{http://www.pythonware.com/products/pil}
 - PIL is Copyright (C) 1997-2006 by Secret Labs AB
 - PIL is Copyright (C) 1995-2006 by Fredrik Lundh

Presentation
============

B{ImageCms} module adds color management features to the PIL library. Even if
little cms provides a python binding, but it is not possible to convert an
entire image from one color space to another. B{ImageCms} allows you to convert
PIL images.

Usage
=====

B{ImageCms} is really easy to use:

>>> import Image, ImageCms
>>> image = Image.open("myPicture.jpg")
>>> sourceProfile = ImageCms.Profile("sRGB.icc")
>>> destinationProfile = ImageCms.Profile("ISOcoated.icc")
>>> transform = ImageCms.Transform(sourceProfile, "RGB", destinationProfile, "CMYK")
>>> imConverted = transform.do(image)
>>> imConverted.save("myPicture.tif")

@author: Frederic Mantegazza
@copyright: 2008 Frederic Mantegazza
@author: Kevin Cazabon (pyCMS)
@copyright: 2002-2003 Kevin Cazabon (pyCMS)
@license: CeCILL-C
@todo: add more features
"""

import os.path

from PIL import Image

import _ImageCms

# Intents

INTENT_PERCEPTUAL = 0
INTENT_RELATIVE_COLORIMETRIC = 1
INTENT_SATURATION = 2
INTENT_ABSOLUTE_COLORIMETRIC = 3

# Flags

FLAGS_MATRIXINPUT            = 0x0001
FLAGS_MATRIXOUTPUT           = 0x0002
FLAGS_MATRIXONLY             = (FLAGS_MATRIXINPUT|FLAGS_MATRIXOUTPUT)

FLAGS_NOWHITEONWHITEFIXUP    = 0x0004    # Don't hot fix scum dot   
FLAGS_NOPRELINEARIZATION     = 0x0010    # Don't create prelinearization tables
                                         # on precalculated transforms (internal use)

FLAGS_GUESSDEVICECLASS       = 0x0020    # Guess device class (for transform2devicelink)

FLAGS_NOTCACHE               = 0x0040    # Inhibit 1-pixel cache

FLAGS_NOTPRECALC             = 0x0100
FLAGS_NULLTRANSFORM          = 0x0200    # Don't transform anyway
FLAGS_HIGHRESPRECALC         = 0x0400    # Use more memory to give better accurancy
FLAGS_LOWRESPRECALC          = 0x0800    # Use less memory to minimize resouces


FLAGS_WHITEBLACKCOMPENSATION = 0x2000
FLAGS_BLACKPOINTCOMPENSATION = FLAGS_WHITEBLACKCOMPENSATION   

# Proofing flags

FLAGS_GAMUTCHECK             = 0x1000    # Out of Gamut alarm
FLAGS_SOFTPROOFING           = 0x4000    # Do softproofing

# Black preservation

FLAGS_PRESERVEBLACK          = 0x8000

# CRD special

FLAGS_NODEFAULTRESOURCEDEF   = 0x01000000

# Gridpoints

def FLAGS_GRIDPOINTS(n):
    return (((n) & 0xFF) << 16)


# How profiles may be used
LCMS_USED_AS_INPUT = 0
LCMS_USED_AS_OUTPUT = 1
LCMS_USED_AS_PROOF = 2


class ImageCmsError(Exception):
    """ General ImageCms error.
    """
    

class Profile(object):
    """ Profile object.
    """
    def __init__(self, filename=None, colorSpace=None, colorTemp=5000):
        """ Init the Profile object.
        """
        self.profile = None
        
        if filename and colorSpace:
            raise ImageCmsError("You can't specify both filename and colorSpace")
        if filename:
            self.__openFromFile(filename)
        elif colorSpace:
            self.colorSpace = colorSpace
            self.colorTemp = colorTemp
            self.__createProfile(colorSpace, colorTemp)
        else:
            raise ImageCmsError("You must specify at least one of filename or colorSpace param")

    def __openFromFile(self, filename):
        """ Load the profile from file.
    
        @param filename: valid filename path to the ICC profile to open
        @type filename: str
    
        @raise ImageCmsError: profileFilename is not a vaild filename for an ICC profile
        """
        if not os.path.isfile(filename):
            raise ImageCmsError("Invalid profile path provided: %s" % filename)
    
        result = _ImageCms.openProfileFromeFile(filename)
    
        if isinstance(result, str):
            raise ImageCmsError(result)
    
        self.profile = result

    def __createProfile(self, colorSpace, colorTemp=5000):
        """ Create a profile.
    
        Use this function to create common profiles on-the-fly instead of
        having to supply a profile on disk and knowing the path to it.
    
        @param colorSpace: color space of the profile to create.
                           (only "LAB", "XYZ", and "sRGB" are supported)
        @type colorSpace: str
    
        @param colorTemp: white point for the profile, in Kelvin (default to D50
                          illuminant, 5000K). colorTemp is ONLY applied to LAB
                          profiles, and is ignored for XYZ and sRGB.
        @type colorTemp: int
    
        @raise ImageCmsError: colorSpace not in ["LAB", "XYZ", "sRGB"];
                           using LAB and colorTemp != positive integer;
                           an error occurs while creating the profile
        """
        if colorSpace not in ["LAB", "XYZ", "sRGB"]:
            raise ImageCmsError("Color space for on-the-fly profile creation must be in ('LAB', 'XYZ', 'sRGB')")
    
        if colorSpace == "LAB":
            if isinstance(colorTemp, int):
                colorTemp = int(colorTemp + 0.5)
            else:
                raise ImageCmsError("Color temperature must be a positive integer, \"%s\" not valid" % colorTemp)
    
        result = _ImageCms.createProfile(colorSpace, colorTemp)
    
        if isinstance(result, str):
            raise ImageCmsError(result)
    
        self.profile = result

    def __getName(self):
        """ Get inner profile name.
    
        this function returns the INTERNAL name of the profile (stored
        in an ICC tag in the profile itself), usually the one used when the
        profile was originally created. Sometimes this tag also contains
        additional information supplied by the creator.
    
        @return: inner profile name (stored in ICC tag)
        @rtype: str
    
        @raise ImageCmsError: profile isn't a valid L{Profile} object;
                           an error occurs while trying to obtain the name tag
        """
        result = _ImageCms.getProfileName(self.profile)
    
        if len(result) > 7 and result[:7] == "ERROR: ":
            raise ImageCmsError(result)
    
        return result

    def __getInfo(self):
        """ Get info from profile.
    
        this function returns the information stored in the profile's
        info tag. This often contains details about the profile, and how it
        was created, as supplied by the creator.
    
        @return: internal profile information (stored in ICC tag)
        @rtype: str
    
        @raise ImageCmsError: profile isn't a valid L{Profile} object;
                           An error occurs while trying to obtain the info tag
        """
        result = _ImageCms.getProfileInfo(self.profile)
    
        if len(result) > 7 and result[:7] == "ERROR: ":
            raise ImageCmsError(result)
    
        return result

    name = property(__getName, doc="Inner profile name")
    info = property(__getInfo, doc="Profile info")

    def getDefaultIntent(self):
        """ Get default profile intent.
    
        Use this function to determine the default (and usually best optomized)
        rendering intent for this profile. Most profiles support multiple
        rendering intents, but are intended mostly for one type of conversion.
        If you wish to use a different intent than returned, use
        isIntentSupported() to verify it will work first.
    
        @return: default rendering intent for this profile
        @rtype: int
    
        @raise ImageCmsError: profile isn't a valid L{Profile} object;
                           An error occurs while trying to obtain the default intent
        """
        result = _ImageCms.getDefaultIntent(self.profile)
    
        if isinstance(result, str):
            raise ImageCmsError(result)
    
        return result
    
    def isIntentSupported(self, intent, direction):
        """ Test if specified intent and direction is available fo that profile.
    
        Use this function to verify that you can use your desired
        intent with profile, and that profile can be used for the
        source/destination/display profile as you desire.
    
        Some profiles are created specifically for one "direction", can cannot
        be used for others. Some profiles can only be used for certain
        rendering intents... so it's best to either verify this before trying
        to create a transform with them (using this function), or catch the
        potential ImageCmsError that will occur if they don't support the modes
        you select.
    
        @return: True if supported, False otherwise
        @rtype: bool
    
        @param intent: rendering intent to use with this profile
        @type intent: int
    
        @param direction: 0 for input, 1 for output, 2 for proof (display)
        @type direction: int
    
        @raise ImageCmsError: profile isn't a valid L{Profile} object;
                           An error occurs while trying to check the intent
        """
        result = _ImageCms.isIntentSupported(self.profile, intent, direction)
    
        if isinstance(result, str):
            raise ImageCmsError(result)
    
        return result


class Transform(object):
    """ Transform object.
    
    Transform object is used to convert images from source color space to
    destination color space. Using a transform rather than using
    profileToProfile() function can save time if the same conversion has to be
    used several times.
    
    If a display profile is given at init time, then a proof tranform is build;
    a proof transform works as a normal transform, but tries to simulate the
    result that would be obtained on the destination profile device using intent
    and display intent to determine what to do with out-of-gamut colors. This is
    known as "soft-proofing". It will ONLY work for converting images that are
    in source mode to images that are in destination mode color format
    (PIL mode, i.e. "RGB", "RGBA", "CMYK", etc.).

    Proof profiling is generally used when using a "proof" device to get a good
    idea of what the final printed/displayed image would look like on the
    destination profile device when it's quicker and easier to use the display
    device for judging color. Generally, this means that display device is a
    monitor, or a dye-sub printer (etc.), and the destination device is
    something more expensive, complicated, or time consuming (making it
    difficult to make a real print for color judgement purposes).

    Soft-proofing basically functions by limiting the color gamut on the
    display device to the gamut availabile on the destination device. However,
    when the final destination device has a much wider gamut than the display
    device, you may obtain marginal results.
    """
    def __init__(self, sourceProfile, sourceMode, destinationProfile, destinationMode, intent=INTENT_PERCEPTUAL,
                 displayProfile=None, displayIntent=INTENT_PERCEPTUAL, flags=0):
        """ Init the Transform object.
    
        @param sourceProfile: ICC source profile to use  for this transform
        @type sourceProfile: L{Profile}
    
        @param sourceMode: valid PIL mode that the appropriate profile also
                           supports (i.e. "RGB", "RGBA", "CMYK", etc.)
        @type sourceMode: str
    
        @param destinationProfile: ICC destination profile to use for this transform
        @type destinationProfile: L{Profile}
    
        @param destinationMode: valid PIL mode that the appropriate profile also
                                supports (i.e. "RGB", "RGBA", "CMYK", etc.)
        @type destinationMode: str
    
        @param intent: rendering intent to use for the transform
        @type intent: int
    
        @param displayProfile: optional ICC display profile to use for this transform
                               (if not None, a proof transform will be build)
        @type displayProfile: L{Profile}
    
        @param displayIntent: rendering intent to use for the source/destination
                              simulation to display transform
        @type displayIntent: int
    
        @param flags: additional flags
        @type flags: int
    
        @return: transform
        @rtype: L{Transform}
    
        @raise ImageCmsError: source, destination or display profiles specified are not
                              valid filenames;
                              an error occurs during creation of the transform;
                              sourceMode or destinationMode are not a mode supported by the
                              destinationProfile (or by ImageCms)
        """
        self.sourceProfile = sourceProfile
        self.sourceMode = sourceMode
        self.destinationMode = destinationMode
        self.destinationProfile = destinationProfile
        self.intent = intent
        self.displayProfile = displayProfile
        self.displayIntent = displayIntent
        self.flags = flags

        if not isinstance(intent, int) or not (0 <= intent <= 3):
            raise ImageCmsError("intent must be int in [0-3]")

        if not isinstance(flags, int):
            raise ImageCmsError("flags must be int")
    
        if displayProfile is not None:
            result = _ImageCms.createProofingTransform(sourceProfile.profile, sourceMode,
                                                       destinationProfile.profile, destinationMode,
                                                       displayProfile.profile,
                                                       intent, displayIntent, flags)
        else:
            result = _ImageCms.createTransform(sourceProfile.profile, sourceMode,
                                               destinationProfile.profile, destinationMode,
                                               intent, flags)
        
        if isinstance(result, str):
            raise ImageCmsError(result)

        self.transform = result
    
    def do(self, image):
        """ Apply the transform to an image.
    
        @param image: an open PIL image object
        @type image: L{Image}
    
        @return: new PIL image
        @rtype: L{Image}
    
        @raise ImageCmsError: imamge.mode != self.sourceMode;
                           image.mode, transfer.sourceMode, or transfer.destinationMode is
                           not supported by _ImageCms or the profiles used for the
                           transform;
                           an error occurs while the transform is being applied
        """
        if image.mode != self.sourceMode:
            raise ImageCmsError("Image mode does not match profile source mode (%s vs %s)"
                                % (image.mode, self.sourceMode))
    
        outputImage = Image.new(self.destinationMode, image.size)
        image.load() # make sure it's loaded, or it may not have an .im attribute!
        
        result = _ImageCms.doTransform(image.im.id, outputImage.im.id, self.transform)
    
        if result == 0:
            return outputImage
        elif result == -1:
            raise ImageCmsError("Error occurred in _ImageCms.applyTransform()")
        else:
            raise ImageCmsError(result)


def versions():
    """ Get versions.

    @return: python, PIL, littlecms and ImageCms versions
    @rtype: dict
    """
    versions = _ImageCms.versions()
    return {'PYTHONVERSION': versions[0],
            'PILVERSION': versions[1],
            'LITTLECMSVERSION': versions[2],
            'PYCMSVERSION': versions[3]}


def profileToProfile(image, sourceProfile, destinationProfile, destinationMode=None, intent=INTENT_PERCEPTUAL, flags=0):
    """ Convert image using source/dest. profiles.

    This function applies an ICC transformation to image from sourceProfile
    color space to destinationProfile color space using the specified rendering
    intent to decide how to handle out-of-gamut colors.

    OutputMode can be used to specify that a color mode conversion is to
    be done using these profiles, but the specified profiles must be able
    to handle that mode. I.e., if converting image from RGB to CMYK using
    profiles, the source profile must handle RGB data, and the destination
    profile must handle CMYK data.

    @param image: an open PIL image object
    @type image: L{Image}

    @param sourceProfile: ICC source profile to use for this image
    @type sourceProfile: L{Profile}

    @param destinationProfile: ICC destination profile to use for this image
    @type destinationProfile: L{Profile}

    @param destinationMode: valid PIL mode that the appropriate profile also
                       supports (i.e. "RGB", "RGBA", "CMYK", etc.)
                       Note: if rendering the image inPlace, destinationMode MUST be
                       the same mode as the source, or omitted completely. If
                       omitted, the destinationMode will be the same as the mode of
                       the input image (image.mode)
    @type destinationMode: str

    @param intent: rendering intent you wish to use for the transform
                            see the ImageCms documentation for details on rendering
                            intents and what they do.
    @type intent: int
    
    @param flags: additional flags
    @type flags: int

    @return: new PIL image
    @rtype: L{Image}

    @raise ImageCmsError: source or destination profiles specified are not valid
                       filenames;
                       inPlace is True and destinationMode != image.mode;
                       an error occurs during application of profiles;
                       destinationMode is not a mode supported by the destinationProfile
                       (or by ImageCms)
    """
    if destinationMode is None:
        destinationMode = image.mode
    if not isinstance(intent, int) or not (0 <= intent <=3):
        raise ImageCmsError("intent must be int in [0-3]")
    if not isinstance(flags, int):
        raise ImageCmsError("flags must be int")

    outputImage = Image.new(destinationMode, image.size)
    image.load() # make sure it's loaded, or it may not have a .im attribute!
    result = _ImageCms.profileToProfile(image.im.id, sourceProfile.profile, outputImage.im.id, destinationProfile.profile, intent, flags)

    if result == 0:
        return outputImage
    elif result == -1:
        raise ImageCmsError("Error occurred in _ImageCms.profileToProfile()")
    else:
        raise ImageCmsError(result)
