/* C low-level functions implementation for ImageCms module

---------------------------------------------------------------------------

Copyright (C) 2008 Frederic Mantegazza
ImageCms home page: http://trac.gbiloba.org/ImageCms

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software. You can use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty and the software's author, the holder of the
economic rights, and the successive licensors have only limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading, using, modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean that it is complicated to manipulate, and that also
therefore means that it is reserved for developers and experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and, more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

---------------------------------------------------------------------------

_ImageCms.c is based on pyCMS code:

pyCMS home page: http://www.cazabon.com
pyCMS is Copyright (C) 2002-2003 Kevin Cazabon

The name has changed to better match other PIL modules. The license has also
changed, moved from LGPL to CeCILL-C to match french laws.

ImageCms also uses little cms and PIL libraries:

little cms home page: http://www.littlecms.com
little cms is Copyright (C) 1998-2008 Marti Maria

PIL home page: http://www.pythonware.com/products/pil
PIL is Copyright (C) 1997-2006 by Secret Labs AB
                 (C) 1995-2006 by Fredrik Lundh
*/

#include "Python.h"
#include "patchlevel.h"
#include "lcms.h"
#include "Imaging.h"

#define PIL_VERSION      "1.1.6"
#define IMAGECMS_VERSION "0.9beta1"

// LCMS_ERROR_SHOW      pop-up window showing error, do not close application
// LCMS_ERROR_ABORT     pop-up window showing error, close the application
// LCMS_ERROR_IGNORE    ignore the error and continue
#define ERROR_LEVEL LCMS_ERROR_SHOW

typedef struct {
    PyObject_HEAD
    Imaging image;
} ImagingObject;

char* errorMsg;


// Helpers
int _errorHandler(int errorCode, const char *errorText)
{
    /*printf("_errorHandler(): errorCode=%d, errorTxt=%s\n", errorCode, errorText);*/
    /*errorMsg = errorText;*/
    errorMsg = errorText;
    
    /*return 0;  // Littlecms prints error message*/
    return -1; // Littlecms does not print error message
}


DWORD _findLcmsType(char* PILmode)
{
    if (strcmp(PILmode, "RGB") == 0) {
        return TYPE_RGBA_8;
    }
    else if (strcmp(PILmode, "RGBA") == 0) {
        return TYPE_RGBA_8;
    }
    else if (strcmp(PILmode, "RGBX") == 0) {
        return TYPE_RGBA_8;
    }
    else if (strcmp(PILmode, "RGBA;16B") == 0) {
        return TYPE_RGBA_16;
    }
    else if (strcmp(PILmode, "CMYK") == 0) {
        return TYPE_CMYK_8;
    }
    else if (strcmp(PILmode, "L") == 0) {
        return TYPE_GRAY_8;
    }
    else if (strcmp(PILmode, "L;16") == 0) {
        return TYPE_GRAY_16;
    }
    else if (strcmp(PILmode, "L;16B") == 0) {
        return TYPE_GRAY_16_SE;
    }
    else if (strcmp(PILmode, "YCCA") == 0) {
        return TYPE_YCbCr_8;
    }
    else if (strcmp(PILmode, "YCC") == 0) {
        return TYPE_YCbCr_8;
    }
    else {
        return TYPE_GRAY_8; // return -1
    }
}


cmsHTRANSFORM _createTransform(cmsHPROFILE hInputProfile, char *sInMode,
                               cmsHPROFILE hOutputProfile, char *sOutMode,
                               int iRenderingIntent, DWORD cmsFLAGS)
{
    cmsHTRANSFORM hTransform;

    cmsErrorAction(ERROR_LEVEL);

    Py_BEGIN_ALLOW_THREADS

    hTransform = cmsCreateTransform(hInputProfile, _findLcmsType(sInMode),
                                    hOutputProfile, _findLcmsType(sOutMode),
                                    iRenderingIntent, cmsFLAGS);

    Py_END_ALLOW_THREADS

    return hTransform;
}


cmsHTRANSFORM _createProofingTransform(cmsHPROFILE hInputProfile, char *sInMode,
                                      cmsHPROFILE hOutputProfile, char *sOutMode,
                                      cmsHPROFILE hDisplayProfile,
                                      int iRenderingIntent, int iDisplayIntent, DWORD cmsFLAGS)
{
    cmsHTRANSFORM hTransform;

    cmsErrorAction(ERROR_LEVEL);

    Py_BEGIN_ALLOW_THREADS

    hTransform = cmsCreateProofingTransform(hInputProfile, _findLcmsType(sInMode),
                                            hOutputProfile, _findLcmsType(sOutMode),
                                            hDisplayProfile,
                                            iRenderingIntent, iDisplayIntent, cmsFLAGS);

    Py_END_ALLOW_THREADS

    return hTransform;
}


int _doTransform(Imaging im, Imaging imOut, cmsHTRANSFORM hTransform) {
    int i;

    if (im->xsize > imOut->xsize) {
        return -1;
    }
    if (im->ysize > imOut->ysize) {
        return -1;
    }

    Py_BEGIN_ALLOW_THREADS

    for (i=0; i < im->ysize; i++) {
        cmsDoTransform(hTransform, im->image[i],
                                   imOut->image[i],
                                   im->xsize);
    }

    Py_END_ALLOW_THREADS

    return 0;
}


// Python functions
static PyObject* versions(PyObject *self, PyObject *args)
{
    return Py_BuildValue("ssss", PY_VERSION, PIL_VERSION, LCMS_VERSION, IMAGECMS_VERSION);
}


static PyObject* profileToProfile(PyObject *self, PyObject *args)
{
    Imaging im;
    Imaging imOut = NULL;
    long idIn;
    long idOut = 0L;
    void *pInputProfile;
    void *pOutputProfile;
    int iRenderingIntent = 0;
		int cmsFLAGS = 0;
    char *inMode;
    char *outMode;
    int result;
    cmsHPROFILE hInputProfile, hOutputProfile;
    cmsHTRANSFORM hTransform;

    // Parse the PyObject arguments, assign to variables accordingly
    if (!PyArg_ParseTuple(args, "lOlO|ii", &idIn, &pInputProfile,
                                          &idOut, &pOutputProfile,
                                          &iRenderingIntent, &cmsFLAGS)) {
        return Py_BuildValue("s", "ERROR: Could not parse the argument tuple passed to _ImageCms.profileToProfile()");
    }

    im = (Imaging) idIn;
    imOut = (Imaging) idOut;

    cmsErrorAction(ERROR_LEVEL);

    inMode = im->mode;
    outMode = imOut->mode;

    hInputProfile = (cmsHPROFILE)PyCObject_AsVoidPtr(pInputProfile);
    hOutputProfile = (cmsHPROFILE)PyCObject_AsVoidPtr(pOutputProfile);

    // Create the transform
    hTransform = _createTransform(hInputProfile, inMode, hOutputProfile, outMode, iRenderingIntent, cmsFLAGS);

    // Apply the transform to imOut
    result = _doTransform(im, imOut, hTransform);

    // Return 0 on success, -1 on failure
    return Py_BuildValue("i", result);
}


static PyObject* openProfileFromeFile(PyObject *self, PyObject *args)
{
    char *sProfile = NULL;
    cmsHPROFILE hProfile;

    if (!PyArg_ParseTuple(args, "s", &sProfile)) {
        return Py_BuildValue("s", "ERROR: Could not parse argument tuple passed to _ImageCms.openProfileFromeFile()");
    }

    cmsErrorAction(ERROR_LEVEL);

    hProfile = cmsOpenProfileFromFile(sProfile, "r");

    return Py_BuildValue("O", PyCObject_FromVoidPtr(hProfile, (void*) cmsCloseProfile));
}


static PyObject* createProfile(PyObject *self, PyObject *args)
{
    char *sColorSpace;
    cmsHPROFILE hProfile;
    int iColorTemp = 0;
    LPcmsCIExyY whitePoint = NULL;
    BOOL result;

    if (!PyArg_ParseTuple(args, "s|i", &sColorSpace, &iColorTemp)) {
        return Py_BuildValue("s", "ERROR: Could not parse the argument tuple passed to _ImageCms.createProfile()");
    }

    cmsErrorAction(ERROR_LEVEL);

    if (strcmp(sColorSpace, "LAB") == 0) {
        result = cmsWhitePointFromTemp(iColorTemp, whitePoint);
        if (result == FALSE) {
            return Py_BuildValue("s", "ERROR: Could not calculate white point from color temperature provided");
        }
        hProfile = cmsCreateLabProfile(whitePoint);
    }
    else if (strcmp(sColorSpace, "XYZ") == 0) {
        hProfile = cmsCreateXYZProfile();
    }
    else if (strcmp(sColorSpace, "sRGB") == 0) {
        hProfile = cmsCreate_sRGBProfile();
    }
    else {
        return Py_BuildValue("s", "ERROR: color space requested is not valid for built-in profiles");
    }

    return Py_BuildValue("O", PyCObject_FromVoidPtr(hProfile, (void*) cmsCloseProfile));
}


static PyObject* createTransform(PyObject *self, PyObject *args)
{
    char *sInMode;
    char *sOutMode;
    int iRenderingIntent = 0;
		int cmsFLAGS = 0;
    void *pInputProfile;
    void *pOutputProfile;
    cmsHPROFILE hInputProfile, hOutputProfile;
    cmsHTRANSFORM htransform;

    if (!PyArg_ParseTuple(args, "OsOs|ii", &pInputProfile, &sInMode,
                                          &pOutputProfile, &sOutMode,
                                          &iRenderingIntent, &cmsFLAGS)) {
        return Py_BuildValue("s", "ERROR: Could not parse argument tuple passed to _ImageCms.createTransformFromOpenProfiles()");
    }

    cmsErrorAction(ERROR_LEVEL);

    hInputProfile = (cmsHPROFILE)PyCObject_AsVoidPtr(pInputProfile);
    hOutputProfile = (cmsHPROFILE)PyCObject_AsVoidPtr(pOutputProfile);

    htransform = _createTransform(hInputProfile, sInMode, hOutputProfile, sOutMode, iRenderingIntent, cmsFLAGS);

    // We don't have to close these profiles... but do we have to decref them?
    
    if (htransform == NULL) {
        return Py_BuildValue("s", errorMsg);
    }
    else {
        return PyCObject_FromVoidPtr(htransform, cmsDeleteTransform);
    }
}


static PyObject* createProofingTransform(PyObject *self, PyObject *args)
{
    char *sInMode;
    char *sOutMode;
    int iRenderingIntent = 0;
    int iDisplayIntent = 0;
		int cmsFLAGS = 0;
    void *pInputProfile;
    void *pOutputProfile;
    void *pDisplayProfile;
    cmsHPROFILE hInputProfile, hOutputProfile, hDisplayProfile;
    cmsHTRANSFORM htransform;

    if (!PyArg_ParseTuple(args, "OsOsO|iii", &pInputProfile, &sInMode,
                                            &pOutputProfile, &sOutMode,
                                            &pDisplayProfile,
                                            &iRenderingIntent, &iDisplayIntent, &cmsFLAGS)) {
        return Py_BuildValue("s", "ERROR: Could not parse argument tuple passed to _ImageCms.createProofingTransform()");
    }

    cmsErrorAction(ERROR_LEVEL);

    hInputProfile = (cmsHPROFILE)PyCObject_AsVoidPtr(pInputProfile);
    hOutputProfile = (cmsHPROFILE)PyCObject_AsVoidPtr(pOutputProfile);
    hDisplayProfile = (cmsHPROFILE)PyCObject_AsVoidPtr(pDisplayProfile);

    htransform = _createProofingTransform(hInputProfile, sInMode,
                                     hOutputProfile, sOutMode,
                                     hDisplayProfile,
                                     iRenderingIntent, iDisplayIntent, cmsFLAGS);

    // We don't have to close these profiles, but do we have to decref them?

    if (htransform == NULL) {
        return Py_BuildValue("s", errorMsg);
    }
    else {
        return PyCObject_FromVoidPtr(htransform, cmsDeleteTransform);
    }
}


static PyObject* doTransform(PyObject *self, PyObject *args)
{
    long idIn;
    long idOut;
    void *hTransformPointer;
    cmsHTRANSFORM hTransform;
    Imaging im;
    Imaging imOut;
    int result;

    if (!PyArg_ParseTuple(args, "llO", &idIn, &idOut, &hTransformPointer)) {
        return Py_BuildValue("s", "ERROR: Could not parse the data passed to _ImageCms.doTransform()");
    }

    im = (Imaging) idIn;
    imOut = (Imaging) idOut;

    cmsErrorAction(ERROR_LEVEL);

    hTransform = (cmsHTRANSFORM)PyCObject_AsVoidPtr(hTransformPointer);

    result = _doTransform(im, imOut, hTransform);

    return Py_BuildValue("i", result);
}


static PyObject* getProfileName(PyObject *self, PyObject *args)
{
    char name[1024];
    void *pProfile;
    cmsHPROFILE hProfile;

    if (!PyArg_ParseTuple(args, "O", &pProfile)) {
        return Py_BuildValue("s", "ERROR: Could not parse the argument tuple passed to _ImageCms.getProfileName()");
    }

    cmsErrorAction(ERROR_LEVEL);

    hProfile = (cmsHPROFILE)PyCObject_AsVoidPtr(pProfile);

    // is there a better way to do this?  I can't seem to work with the const char* return value otherwise
    sprintf(name, "%s", cmsTakeProductName(hProfile));

    return Py_BuildValue("s", name);
}


static PyObject* getProfileInfo(PyObject *self, PyObject *args)
{
    char info[4096];
    void *pProfile;
    cmsHPROFILE hProfile;

    if (!PyArg_ParseTuple(args, "O", &pProfile)) {
        return Py_BuildValue("s", "ERROR: Could not parse the argument tuple passed to _ImageCms.getProfileInfo()");
    }

    cmsErrorAction(ERROR_LEVEL);

    hProfile = (cmsHPROFILE)PyCObject_AsVoidPtr(pProfile);

    // is there a better way to do this?  I can't seem to work with the const char* return value otherwise
    sprintf(info, "%s", cmsTakeProductInfo(hProfile));

    return Py_BuildValue("s", info);
}


static PyObject* getDefaultIntent(PyObject *self, PyObject *args)
{
    int intent = 0;
    void *pProfile;
    cmsHPROFILE hProfile;

    if (!PyArg_ParseTuple(args, "O", &pProfile)) {
        return Py_BuildValue("s", "ERROR: Could not parse the argument tuple passed to _ImageCms.getDefaultIntent()");
    }

    cmsErrorAction(ERROR_LEVEL);

    hProfile = (cmsHPROFILE)PyCObject_AsVoidPtr(pProfile);

    intent = cmsTakeRenderingIntent(hProfile);

    return Py_BuildValue("i", intent);
}


static PyObject* isIntentSupported(PyObject *self, PyObject *args)
{
    int iIntent;
    int iDirection;
    void *pProfile;
    BOOL result;
    cmsHPROFILE hProfile;

    if (!PyArg_ParseTuple(args, "Oii", &pProfile, &iIntent, &iDirection)) {
        return Py_BuildValue("s", "ERROR: Could not parse the argument tuple passed to _ImageCms.isIntentSupported()");
    }

    cmsErrorAction(ERROR_LEVEL);

    hProfile = (cmsHPROFILE)PyCObject_AsVoidPtr(pProfile);

    result = cmsIsIntentSupported(hProfile, iIntent, iDirection);

    if (result == TRUE) {
        return Py_BuildValue("i", 1);
    }
    else {
        return Py_BuildValue("i", -1);
    }
}


// Python interface setup
static PyMethodDef _ImageCms_methods[] = {

    // ImageCms info
    {"versions", versions, 1, \
"versions()\n\n \
returns tuple of ImageCmsversion, littleCMSversion, pythonVersion that it was compiled with.\n \
Note: don't trust this 100%, they must be set manually in the source code for now)"
    },

    // profile and transform functions
    {"profileToProfile", profileToProfile, 1, \
"profileToProfile(idIn, idOut, InputProfile, OutputProfile, [RenderingIntent], [Flags])\n\n \
returns 0 on success, -1 on failure.\n \
If idOut is the same as idIn, idIn is modified in place, otherwise the results\n \
are applied to idOut"
    },
    {"openProfileFromeFile", openProfileFromeFile, 1, \
"openProfileFromeFile(profileName)\n\n \
returns a handle to an open ImageCms profile that can be used to build a transform"},
    {"createProfile", createProfile, 1, \
"_ImageCms.createProfile(colorSpace, [colorTemp])\n\n \
returns a handle to an open profile created on the fly.\n \
colorSpace can be 'LAB', 'XYZ', or 'xRGB'.\n \
If using LAB, you can specify a white point color temperature,\n \
or let it default to D50 (5000K)"
    },
    {"createTransform", createTransform, 1, \
"createTransform(InputProfile, InMode, OutputProfile, OutMode, [RenderingIntent], [Flags])\n\n \
returns a handle to a pre-computed ICC transform that can be used for processing\n \
multiple images, saving calculation time"
    },
    {"createProofingTransform", createProofingTransform, 1, \
"createProofingTransform(InputProfile, InMode, OutputProfile, OutMode, DisplayProfile, [RenderingIntent], [DisplayRenderingIntent], [Flags])\n\n \
returns a handle to a pre-computed soft-proofing (simulating the output device\n \
capabilities on the display device)ICC transform that can be used for processing\n \
multiple images, saving calculation time"
    },
    {"doTransform", doTransform, 1, \
"doTransform(idIn, idOut, hTransform)\n\n \
applies a pre-calcuated transform (from _ImageCms.createTransform) to an image.\n \
If idIn and idOut are the same, it modifies the image in place, otherwise the\n \
new image is built in idOut.  Returns 0 on success, -1 on failure"
    },
    
    // on-the-fly profile creation functions

    // profile info functions
    {"getProfileName", getProfileName, 1, \
"getProfileName(profile)\n\n \
returns the internal name of the profile"
    },
    {"getProfileInfo", getProfileInfo, 1, \
"getProfileInfo(profile)\n\n \
returns additional information about the profile"
    },
    {"getDefaultIntent", getDefaultIntent, 1, \
"getDefaultIntent(profile)\n\n \
returns the default rendering intent of the profile (as an integer)"
    },
    {"isIntentSupported", isIntentSupported, 1, \
"isIntentSupported(profile, intent, direction)\n\n \
returns 1 if profile supports that intent, -1 if it doesnt.\n \
Direction is what the profile is being used for: 0 (input), 1 (output), 2 (proof)"
    },

    {NULL, NULL}
};


void init_ImageCms(void)
{
    Py_InitModule("_ImageCms", _ImageCms_methods);
    cmsSetErrorHandler(_errorHandler);
}
